/*
	SObjectizer 5.
*/

#include <so_5/rt/h/disp.hpp>

#include <so_5/h/exception.hpp>

namespace so_5
{

//
// dispatcher_t
//

void
dispatcher_t::set_data_sources_name_base(
	const std::string & /*name*/ )
{
}

} /* namespace so_5 */
