/*
 * SObjectizer-5
 */

/*!
 * \file
 * \since
 * v.5.5.22
 *
 * \brief Definition of optional type.
 */

#pragma once

#include <so_5/3rd_party/optional-lite/nonstd/optional.hpp>

namespace so_5 {

using nonstd::optional;

} /* namespace so_5 */

