/*
 * SObjectizer-5
 */

/*!
 * \since
 * v.5.4.0
 *
 * \file
 * \brief A helper header file for including all public SObjectizer stuff.
 */

#pragma once

#include <so_5/rt/h/rt.hpp>
#include <so_5/api/h/api.hpp>
#include <so_5/h/wrapped_env.hpp>
#include <so_5/rt/h/env_infrastructures.hpp>

#include <so_5/rt/h/enveloped_msg.hpp>

#include <so_5/h/mchain_helper_functions.hpp>
#include <so_5/h/thread_helper_functions.hpp>

#include <so_5/disp/one_thread/h/pub.hpp>
#include <so_5/disp/active_obj/h/pub.hpp>
#include <so_5/disp/active_group/h/pub.hpp>
#include <so_5/disp/thread_pool/h/pub.hpp>
#include <so_5/disp/adv_thread_pool/h/pub.hpp>
#include <so_5/disp/prio_one_thread/strictly_ordered/h/pub.hpp>
#include <so_5/disp/prio_one_thread/quoted_round_robin/h/pub.hpp>
#include <so_5/disp/prio_dedicated_threads/one_per_prio/h/pub.hpp>

#include <so_5/h/version.hpp>

