/*
	SObjectizer 5.
*/

/*!
	\file
	\brief A wrapper for all major SO Run-Time includes.
*/

#pragma once

#include <so_5/rt/h/message.hpp>
#include <so_5/rt/h/agent.hpp>
#include <so_5/rt/h/agent_coop.hpp>
#include <so_5/rt/h/state.hpp>
#include <so_5/rt/h/event_data.hpp>
#include <so_5/rt/h/mbox.hpp>
#include <so_5/rt/h/environment.hpp>
#include <so_5/rt/h/agent_coop_notifications.hpp>
#include <so_5/rt/h/send_functions.hpp>

#include <so_5/rt/h/mchain_select.hpp>

#include <so_5/h/chrono_helpers.hpp>

#include <so_5/rt/stats/h/std_names.hpp>
#include <so_5/rt/stats/h/messages.hpp>

#include <so_5/rt/h/tuple_as_message.hpp>
